import pygame
import math
import random

import utilities
import level
import animatedSprite8
import animatedSprite16
import animatedSprite24
import musicManager
import soundFxManager
import hiscoreManager
import bonusObjectManager
import pillManager

GAME_VERSION_STRING = "V1.4"

RANDOM_SEED = 1337

PLAYER_STATE_WALK_LEFT = 0
PLAYER_STATE_WALK_RIGHT = 1
PLAYER_STATE_JUMPING = 2
PLAYER_STATE_FLOAT_LEFT = 3
PLAYER_STATE_FLOAT_RIGHT = 4
PLAYER_STATE_IDLE_LEFT = 5
PLAYER_STATE_IDLE_RIGHT = 6
PLAYER_STATE_DYING = 7

PLAYER_ANIM_HOLD_DURATION = 10

FACING_LEFT = 0
FACING_RIGHT = 1

PLAYER_HORZ_SPEED_PER_FRAME = 1.0
PLAYER_JUMP_SPEED_PER_FRAME = 4.0
PLAYER_FLOAT_SPEED_PER_FRAME = 0.25
PLAYER_FALL_SPEED_PER_FRAME = 1.0

KEY_JUMP_FLOAT = pygame.K_UP
KEY_MOVE_DROP = pygame.K_DOWN
KEY_MOVE_LEFT = pygame.K_LEFT
KEY_MOVE_RIGHT = pygame.K_RIGHT
KEY_FIRE = pygame.K_a
KEY_PAUSE_TOGGLE = pygame.K_RETURN
KEY_PAUSE_RETURN_TO_TITLE1 = pygame.K_LSHIFT
KEY_PAUSE_RETURN_TO_TITLE2 = pygame.K_RSHIFT

GP_BACK1 = 6
GP_BACK2 = 8
GP_START1 = 7
GP_START2 = 9
GP_A = 0
GP_B = 1
GP_X = 2
GP_Y = 3

UPWARD_SMOKE_PUFF_HOLD_DURATION = 1
EXPLOSION_HOLD_DURATION = 1

LEVEL_CHANGE_WAIT_TIME = 1500

GET_READY_WAIT_FRAMES = 180
NEXT_LEVEL_WAIT_FRAMES = 180

GAME_STATE_NONE = 0
GAME_STATE_TITLE = 1
GAME_STATE_HISCORES = 2
GAME_STATE_SETTINGS = 3
GAME_STATE_INGAME = 4

PLAYER_SHOT_SPEED = 4
PLAYER_SHOTS_HOLD_DURATION = 1
PLAYER_SHOTS_LIFETIME = 60
PLAYER_SHOT_INTERVAL = 60

ENEMY_SPAWNER_HOLD_DURATION = 2
ENEMY_SPAWNER_LIFETIME = 60

ENEMY_TYPE_NOS = 4 # gun turret type counts once for both ceiling and floor
ENEMY_TYPE_HOMING_MISSILE = 0
ENEMY_TYPE_SAW_BLADE = 1
ENEMY_TYPE_FLOOR_GUN_TURRET = 2
ENEMY_TYPE_CEILING_GUN_TURRET = 3
ENEMY_TYPE_TANK = 4
ENEMY_SPAWN_FREQUENCY = (20.0 * 60.0) # 5 secs baseline
ENEMY_SPAWN_FACTOR = 0.25
ENEMY_SPAWN_MIN_FREQUENCY = 15.0 # 1/4 secs min spawning interval
MAX_ENEMIES_ALIVE = 32

ENEMY_HOMING_MISSILE_HOLD_DURATION = 15
ENEMY_SAW_BLADE_HOLD_DURATION = 5
ENEMY_FLOOR_GUN_TURRET_HOLD_DURATION = 30
ENEMY_CEILING_GUN_TURRET_HOLD_DURATION = 30
ENEMY_TANK_HOLD_DURATION = 15

ENEMY_HOMING_MISSILE_SPEED = 0.5
ENEMY_SAW_BLADE_SPEED = 0.5
ENEMY_TANK_SPEED = 0.5

ENEMY_SENSOR_RANGE = (5 * 16)
ENEMY_DETONATE_DISTANCE = 8
ENEMY_SHOT_INTERVAL = 120
ENEMY_MIN_SHOT_INTERVAL = 30
ENEMY_SHOTS_HOLD_DURATION = 1
ENEMY_SHOTS_SPEED = 1

ENEMY_HIT_DISTANCE = 16
PLAYER_HIT_DISTANCE = 16

#
#
#
class App:
	def __init__(self, newScore, newLives, newLevel):
		random.seed(RANDOM_SEED)

		self.initialScore = newScore
		self.initialLives = newLives
		self.initialLevel = newLevel

		self.score = newScore
		self.lives = newLives
		self.level = newLevel

		self.restartIngameLevel()

		self.prevGameState = GAME_STATE_NONE
		self.gameState = GAME_STATE_NONE
		self.enterGameState(GAME_STATE_TITLE)

		self.frameCounter60Hz = 0

		self.titleMenuButtonIndex = 0

		self.hiscoreManager = hiscoreManager.HiscoreManager()

		self.settingsMenuButtonIndex = 0
		self.soundfxOn = True

		self.musicManager = musicManager.MusicManager()
		self.soundFxManager = soundFxManager.SoundFxManager()

		self.bonusObjectManager = bonusObjectManager.BonusObjectManager(self)
		self.pillManager = pillManager.PillManager(self)

	#
	#
	#
	def enterGameState(self, newGameState):
		if newGameState == self.gameState:
			return

		self.exitGameState()
		self.gameState = newGameState

		if self.gameState == GAME_STATE_TITLE:
			utilities.loadTitleBgToDisplay()

		elif self.gameState == GAME_STATE_HISCORES:
			utilities.loadHiscoreBgToDisplay()

		elif self.gameState == GAME_STATE_SETTINGS:
			utilities.loadSettingsBgToDisplay()

		elif self.gameState == GAME_STATE_INGAME:
			self.score = self.initialScore
			self.lives = self.initialLives
			self.level = self.initialLevel

			self.restartIngameLevel()

			self.hiscoreManager.resetHiscores()

			self.enterHiscoreMode = False

			self.playerShotSpeedModifier = 0.5
			self.playerShotLifetimeModifier = 0.5
			self.playerShotIntervalModifier = 0.5
			self.playerShotIntervalCounter = PLAYER_SHOT_INTERVAL / self.playerShotIntervalModifier # allow immediate shooting

	#
	#
	#
	def exitGameState(self):
		self.prevGameState = self.gameState

		if self.prevGameState == GAME_STATE_TITLE:
			pass

		elif self.prevGameState == GAME_STATE_HISCORES:
			pass

		elif self.prevGameState == GAME_STATE_SETTINGS:
			pass

		elif self.prevGameState == GAME_STATE_INGAME:
			pass

	#
	#
	#
	def handleEvent(self, event):
		# if event.type == pygame.JOYBUTTONDOWN:
		# 		print("Joystick button pressed." + str(event))
		# elif event.type == pygame.JOYBUTTONUP:
		# 	print("Joystick button released." + str(event))
		# elif event.type == pygame.JOYAXISMOTION:
		# 	print("Joystick axis motion." + str(event))
		# elif event.type == pygame.JOYBALLMOTION:
		# 	print("Joystick ball motion." + str(event))
		# elif event.type == pygame.JOYHATMOTION:
		# 	print("Joystick joyhat motion." + str(event))

		if event.type == pygame.JOYBUTTONUP:
			if event.button == GP_BACK1 or event.button == GP_BACK2:
				self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BUTTON_PUSH, 0)
				self.storeGameDataForTermination()
				return True
			elif event.button == GP_START1 or event.button == GP_START2:
				if self.getReadyTime == False and self.nextLevelTime == False:
					self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BUTTON_PUSH, 0)
					self.isPausedGame = not self.isPausedGame

		elif event.type == pygame.KEYUP:
			if event.key == pygame.K_ESCAPE:
				self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BUTTON_PUSH, 0)
				self.storeGameDataForTermination()
				return True

			elif event.key == KEY_PAUSE_TOGGLE:
				if self.getReadyTime == False and self.nextLevelTime == False:
					self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BUTTON_PUSH, 0)
					self.isPausedGame = not self.isPausedGame

		# GameState cheats
		# if event.type == pygame.KEYUP:
		# 	if self.getReadyTime == False and self.nextLevelTime == False:
		# 		if self.isPausedGame == False:
		# 			if event.key == pygame.K_1:
		# 				self.enterGameState(GAME_STATE_TITLE)
		# 			elif event.key == pygame.K_2:
		# 				self.enterGameState(GAME_STATE_HISCORES)
		# 			elif event.key == pygame.K_3:
		# 				self.enterGameState(GAME_STATE_SETTINGS)
		# 			elif event.key == pygame.K_4:
		# 				self.enterGameState(GAME_STATE_INGAME)

		if self.gameState == GAME_STATE_TITLE:
			if event.type == pygame.JOYAXISMOTION:
				if event.axis == 1 and event.value <= -0.5:
					self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BUTTON_CLICK, 0)

					self.titleMenuButtonIndex -= 1
					if self.titleMenuButtonIndex < 0:
						self.titleMenuButtonIndex = 3

				elif event.axis == 1 and event.value >= 0.5:
					self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BUTTON_CLICK, 0)

					self.titleMenuButtonIndex += 1
					if self.titleMenuButtonIndex > 3:
						self.titleMenuButtonIndex = 0

			elif event.type == pygame.JOYHATMOTION:
				if event.value[1] == 1:
 					self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BUTTON_CLICK, 0)

					self.titleMenuButtonIndex -= 1
					if self.titleMenuButtonIndex < 0:
						self.titleMenuButtonIndex = 3

				elif event.value[1] == -1:
					self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BUTTON_CLICK, 0)

					self.titleMenuButtonIndex += 1
					if self.titleMenuButtonIndex > 3:
						self.titleMenuButtonIndex = 0

			elif event.type == pygame.JOYBUTTONUP:
				if event.button == GP_A or event.button == GP_B or event.button == GP_X or event.button == GP_Y:
					self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BUTTON_PUSH, 0)

					if self.titleMenuButtonIndex == 0:
						self.enterGameState(GAME_STATE_INGAME)
					elif self.titleMenuButtonIndex == 1:
						self.enterGameState(GAME_STATE_HISCORES)
					elif self.titleMenuButtonIndex == 2:
						self.enterGameState(GAME_STATE_SETTINGS)
					elif self.titleMenuButtonIndex == 3: # Quit button
						self.storeGameDataForTermination()
						return True

			elif event.type == pygame.KEYUP:
				if event.key == KEY_JUMP_FLOAT:
					self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BUTTON_CLICK, 0)

					self.titleMenuButtonIndex -= 1
					if self.titleMenuButtonIndex < 0:
						self.titleMenuButtonIndex = 3

				elif event.key == KEY_MOVE_DROP:
					self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BUTTON_CLICK, 0)

					self.titleMenuButtonIndex += 1
					if self.titleMenuButtonIndex > 3:
						self.titleMenuButtonIndex = 0

				elif event.key == KEY_FIRE:
					self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BUTTON_PUSH, 0)

					if self.titleMenuButtonIndex == 0:
						self.enterGameState(GAME_STATE_INGAME)
					elif self.titleMenuButtonIndex == 1:
						self.enterGameState(GAME_STATE_HISCORES)
					elif self.titleMenuButtonIndex == 2:
						self.enterGameState(GAME_STATE_SETTINGS)
					elif self.titleMenuButtonIndex == 3: # Quit button
						self.storeGameDataForTermination()
						return True

		elif self.gameState == GAME_STATE_HISCORES:
			if event.type == pygame.JOYBUTTONUP:
				if event.button == GP_START1 or event.button == GP_START2:
					self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BUTTON_PUSH, 0)
					self.enterGameState(GAME_STATE_TITLE)
					return False

			elif event.type == pygame.KEYUP:
				if event.key == KEY_PAUSE_RETURN_TO_TITLE1 or event.key == KEY_PAUSE_RETURN_TO_TITLE2:
					self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BUTTON_PUSH, 0)
					self.enterGameState(GAME_STATE_TITLE)
					return False

		elif self.gameState == GAME_STATE_SETTINGS:
			if event.type == pygame.JOYAXISMOTION:
				if event.axis == 0 and event.value <= -0.5:
					self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BUTTON_CLICK, 0)
					self.settingsMenuButtonIndex -= 1
					if self.settingsMenuButtonIndex < 0:
						self.settingsMenuButtonIndex = 1
				elif event.axis == 0 and event.value >= 0.5:
					self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BUTTON_CLICK, 0)
					self.settingsMenuButtonIndex += 1
					if self.settingsMenuButtonIndex > 1:
						self.settingsMenuButtonIndex = 0

			elif event.type == pygame.JOYHATMOTION:
				if event.value[0] == -1 or event.value[1] == 1:
					self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BUTTON_CLICK, 0)
					self.settingsMenuButtonIndex -= 1
					if self.settingsMenuButtonIndex < 0:
						self.settingsMenuButtonIndex = 1
				elif event.value[0] == 1 or event.value[1] == -1:
					self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BUTTON_CLICK, 0)
					self.settingsMenuButtonIndex += 1
					if self.settingsMenuButtonIndex > 1:
						self.settingsMenuButtonIndex = 0

			elif event.type == pygame.JOYBUTTONUP:
				if event.button == GP_START1 or event.button == GP_START2:
					self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BUTTON_PUSH, 0)
					self.enterGameState(GAME_STATE_TITLE)
					return False
				elif event.button == GP_A or event.button == GP_B or event.button == GP_X or event.button == GP_Y:
					if self.settingsMenuButtonIndex == 0:
						self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BUTTON_PUSH, 0)
						playFlag = self.soundFxManager.getPlaySoundFx()
						self.soundFxManager.playSoundFx(not playFlag)
					elif self.settingsMenuButtonIndex == 1:
						self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BUTTON_PUSH, 0)
						playFlag = self.musicManager.getPlayMusic()
						self.musicManager.playMusic(not playFlag)

			elif event.type == pygame.KEYUP:
				if event.key == KEY_PAUSE_RETURN_TO_TITLE1 or event.key == KEY_PAUSE_RETURN_TO_TITLE2:
					self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BUTTON_PUSH, 0)
					self.enterGameState(GAME_STATE_TITLE)
					return False
				elif event.key == KEY_JUMP_FLOAT or event.key == KEY_MOVE_LEFT:
					self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BUTTON_CLICK, 0)
					self.settingsMenuButtonIndex -= 1
					if self.settingsMenuButtonIndex < 0:
						self.settingsMenuButtonIndex = 1
				elif event.key == KEY_MOVE_DROP  or event.key == KEY_MOVE_RIGHT:
					self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BUTTON_CLICK, 0)
					self.settingsMenuButtonIndex += 1
					if self.settingsMenuButtonIndex > 1:
						self.settingsMenuButtonIndex = 0
				elif event.key == KEY_FIRE:
					if self.settingsMenuButtonIndex == 0:
						self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BUTTON_PUSH, 0)
						playFlag = self.soundFxManager.getPlaySoundFx()
						self.soundFxManager.playSoundFx(not playFlag)
					elif self.settingsMenuButtonIndex == 1:
						self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BUTTON_PUSH, 0)
						playFlag = self.musicManager.getPlayMusic()
						self.musicManager.playMusic(not playFlag)

		elif self.gameState == GAME_STATE_INGAME and self.isPausedGame == False:
			if self.enterHiscoreMode == True:
				if event.type == pygame.JOYAXISMOTION:
					if event.axis == 0 and event.value <= -0.5:
						self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BUTTON_CLICK, 0)

						self.hiscoreManager.setInitialAndMoveToPrev()

					elif event.axis == 0 and event.value >= 0.5:
						self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BUTTON_CLICK, 0)

						self.hiscoreManager.setInitialAndMoveToNext()
				elif event.type == pygame.JOYHATMOTION:
					if event.value[1] == 1:
						self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BUTTON_CLICK, 0)

						self.hiscoreManager.setInitialAndMoveToPrev()

					elif event.value[1] == -1:
						self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BUTTON_CLICK, 0)

						self.hiscoreManager.setInitialAndMoveToNext()

				elif event.type == pygame.JOYBUTTONUP:
					if event.button == GP_A or event.button == GP_B or event.button == GP_X or event.button == GP_Y:
						self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BUTTON_PUSH, 0)

						if self.hiscoreManager.confirmInitialsOrMoveToNext(self.score) == True:
							self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_THX_HERO, 1)
							self.enterGameState(GAME_STATE_HISCORES)

				elif event.type == pygame.KEYUP:
					if event.key == KEY_JUMP_FLOAT:
						self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BUTTON_CLICK, 0)

						self.hiscoreManager.setInitialAndMoveToPrev()

					elif event.key == KEY_MOVE_DROP:
						self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BUTTON_CLICK, 0)

						self.hiscoreManager.setInitialAndMoveToNext()

					elif event.key == KEY_FIRE:
						self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BUTTON_PUSH, 0)

						if self.hiscoreManager.confirmInitialsOrMoveToNext(self.score) == True:
							self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_THX_HERO, 1)
							self.enterGameState(GAME_STATE_HISCORES)

			else:
				if event.type == pygame.JOYAXISMOTION:
					if self.getReadyTime == True:
    						return False

					if self.nextLevelTime == True:
						return False

					if event.axis == 1 and event.value <= -0.5 and self.keyJumpFloatDown == False:
						self.keyJumpFloatStart()
					elif event.axis == 1 and event.value > -0.5 and self.keyJumpFloatDown == True:
						self.keyJumpFloatEnd()

					if event.axis == 1 and event.value >= 0.5 and self.keyDropDown == False:
						self.keyDropStart()
					elif event.axis == 1 and event.value < 0.5 and self.keyDropDown == True:
						self.keyDropEnd()

					if event.axis == 0 and event.value <= -0.5 and self.keyMoveLeftDown == False:
						self.keyMoveLeftStart()
					elif event.axis == 0 and event.value > -0.5 and self.keyMoveLeftDown == True:
						self.keyMoveLeftEnd()

					if event.axis == 0 and event.value >= 0.5 and self.keyMoveRightDown == False:
						self.keyMoveRightStart()
					elif event.axis == 0 and event.value < 0.5 and self.keyMoveRightDown == True:
						self.keyMoveRightEnd()

				elif event.type == pygame.JOYHATMOTION:
					if self.getReadyTime == True:
						return False

					if self.nextLevelTime == True:
						return False

					if event.value[1] == 1 and self.keyJumpFloatDown == False:
						self.keyJumpFloatStart()
					elif event.value[1] != 1 and self.keyJumpFloatDown == True:
						self.keyJumpFloatEnd()

					if event.value[1] == -1 and self.keyDropDown == False:
						self.keyDropStart()
					elif event.value[1] != -1 and self.keyDropDown == True:
						self.keyDropEnd()

					if event.value[0] == -1 and self.keyMoveLeftDown == False:
						self.keyMoveLeftStart()
					elif event.value[0] != -1 and self.keyMoveLeftDown == True:
						self.keyMoveLeftEnd()

					if event.value[0] == 1 and self.keyMoveRightDown == False:
						self.keyMoveRightStart()
					elif event.value[0] != 1 and self.keyMoveRightDown == True:
						self.keyMoveRightEnd()

				elif event.type == pygame.JOYBUTTONDOWN:
					if self.getReadyTime == True:
						return False

					if self.nextLevelTime == True:
						return False

					if event.button == GP_A or event.button == GP_B or event.button == GP_X or event.button == GP_Y:
						self.keyFireStart()

				elif event.type == pygame.KEYDOWN:
					if self.getReadyTime == True:
						return False

					if self.nextLevelTime == True:
						return False

					if event.key == KEY_JUMP_FLOAT:
						self.keyJumpFloatStart()
					elif event.key == KEY_MOVE_DROP:
						self.keyDropStart()
					elif event.key == KEY_MOVE_LEFT:
						self.keyMoveLeftStart()
					elif event.key == KEY_MOVE_RIGHT:
						self.keyMoveRightStart()
					elif event.key == KEY_FIRE:
						self.keyFireStart()

				elif event.type == pygame.JOYBUTTONUP:
					if self.getReadyTime == True:
						return False

					if self.nextLevelTime == True:
						return False

					if event.button == GP_BACK1 or event.button == GP_BACK2:
						self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BUTTON_PUSH, 0)
						self.enterGameState(GAME_STATE_TITLE)
						return False

					if event.button == GP_A or event.button == GP_B or event.button == GP_X or event.button == GP_Y:
						self.keyFireEnd()

				elif event.type == pygame.KEYUP:
					if self.getReadyTime == True:
						return False

					if self.nextLevelTime == True:
						return False

					if event.key == KEY_PAUSE_RETURN_TO_TITLE1 or event.key == KEY_PAUSE_RETURN_TO_TITLE2:
						self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BUTTON_PUSH, 0)
						self.enterGameState(GAME_STATE_TITLE)
						return False

					if event.key == KEY_JUMP_FLOAT:
						self.keyJumpFloatEnd()
					elif event.key == KEY_MOVE_DROP:
						self.keyDropEnd()
					elif event.key == KEY_MOVE_LEFT:
						self.keyMoveLeftEnd()
					elif event.key == KEY_MOVE_RIGHT:
						self.keyMoveRightEnd()
					elif event.key == KEY_FIRE:
						self.keyFireEnd()

					# cheat keys
					# elif event.key == pygame.K_p:
					# 	self.markMoveToNextLevel()
					# elif event.key == pygame.K_l:
					# 	self.markMoveToPrevLevel()

		return False

	#
	#
	#
	def update(self):
		self.musicManager.update()
		self.soundFxManager.update()

		if self.gameState == GAME_STATE_TITLE:
			pass

		elif self.gameState == GAME_STATE_HISCORES:
			pass

		elif self.gameState == GAME_STATE_SETTINGS:
			pass

		elif self.gameState == GAME_STATE_INGAME:
			if self.getReadyTime == True:
				self.getReadyFrameCounter += 1
				if self.getReadyFrameCounter >= GET_READY_WAIT_FRAMES:
					self.getReadyTime = False

			if self.nextLevelTime == True:
				self.nextLevelFrameCounter += 1
				if self.nextLevelFrameCounter >= NEXT_LEVEL_WAIT_FRAMES:
					self.nextLevelTime = False
			else:
				if 	self.moveLevelUp == True:
					self.moveToNextLevel()
				elif self.moveLevelDown == True:
					self.moveToPrevLevel()

				if self.startNewLevel == True:
					self.startNewLevel = False
					self.prepareLevel()

					self.getReadyTime = True
					self.getReadyFrameCounter = 0

				if self.isPausedGame == False:
					self.movePlayer()

			if self.keyFireDown == True:
				if self.playerLastFacing == FACING_LEFT:
					shotSpawnPosition = (self.playerPosition[0] - 8 + PLAYER_SHOT_SPEED, self.playerPosition[1] + 8)
				else:
					shotSpawnPosition = (self.playerPosition[0] + 24 - PLAYER_SHOT_SPEED, self.playerPosition[1] + 8)
				self.addPlayerShot(shotSpawnPosition, self.playerLastFacing)

			self.pillManager.updatePills()
			self.bonusObjectManager.updateBonusObjects()
			self.updateUpdwardSmokePuffs()
			self.updateExplosions()
			self.updatePlayerShots()
			self.updateEnemyShots()
			self.updateEnemySpawners()
			self.updateEnemies()

			self.updatePlayer()

		elif self.gameState == GAME_STATE_PAUSED:
			pass

		self.frameCounter60Hz += 1
		if self.frameCounter60Hz >= 60:
			self.frameCounter60Hz = 0
	#
	#
	#
	def render(self):
		offset = self.frameCounter60Hz / 30

		if self.gameState == GAME_STATE_TITLE:
			utilities.drawTextToDisplay(GAME_VERSION_STRING, (8, 220), utilities.TEXT_COLOR_GREEN)

			if self.titleMenuButtonIndex == 0:
				utilities.blitSignsToDisplaySurface((128, 120), (0, 144 + 22, 64, 22))
				utilities.blitSignsToDisplaySurface((128, 120), (256, 188 + (offset * 22), 64, 22))

				utilities.blitSignsToDisplaySurface((128, 120 + (1 *23)), (128, 144, 64, 22))
				utilities.blitSignsToDisplaySurface((128, 120 + (2 *23)), (256, 144, 64, 22))
				utilities.blitSignsToDisplaySurface((128, 120 + (3 *23)), (64, 144, 64, 22))

			elif self.titleMenuButtonIndex == 1:
				utilities.blitSignsToDisplaySurface((128, 120 + (1 *23)), (128, 144 + 22, 64, 22))
				utilities.blitSignsToDisplaySurface((128, 120 + (1 *23)), (256, 188 + (offset * 22), 64, 22))

				utilities.blitSignsToDisplaySurface((128, 120), (0, 144, 64, 22))
				utilities.blitSignsToDisplaySurface((128, 120 + (2 *23)), (256, 144, 64, 22))
				utilities.blitSignsToDisplaySurface((128, 120 + (3 *23)), (64, 144, 64, 22))

			elif self.titleMenuButtonIndex == 2:
				utilities.blitSignsToDisplaySurface((128, 120 + (2 *23)), (256, 144 + 22, 64, 22))
				utilities.blitSignsToDisplaySurface((128, 120 + (2 *23)), (256, 188 + (offset * 22), 64, 22))

				utilities.blitSignsToDisplaySurface((128, 120), (0, 144, 64, 22))
				utilities.blitSignsToDisplaySurface((128, 120 + (1 *23)), (128, 144, 64, 22))
				utilities.blitSignsToDisplaySurface((128, 120 + (3 *23)), (64, 144, 64, 22))

			elif self.titleMenuButtonIndex == 3:
				utilities.blitSignsToDisplaySurface((128, 120 + (3 *23)), (64, 144 + 22, 64, 22))
				utilities.blitSignsToDisplaySurface((128, 120 + (3 *23)), (256, 188 + (offset * 22), 64, 22))

				utilities.blitSignsToDisplaySurface((128, 120), (0, 144, 64, 22))
				utilities.blitSignsToDisplaySurface((128, 120 + (1 *23)), (128, 144, 64, 22))
				utilities.blitSignsToDisplaySurface((128, 120 + (2 *23)), (256, 144, 64, 22))


		elif self.gameState == GAME_STATE_HISCORES:
			utilities.blitSignsToDisplaySurface((120, 30), (192, 0 + (offset * 16), 80, 16))

			self.hiscoreManager.render()

		elif self.gameState == GAME_STATE_SETTINGS:
			if self.soundFxManager.getPlaySoundFx() == True:
				sfxH_Offset = 128
			else:
				sfxH_Offset = 0

			if self.musicManager.getPlayMusic() == True:
				musicH_Offset = 128
			else:
				musicH_Offset = 0

			if self.settingsMenuButtonIndex == 0:
				utilities.blitSignsToDisplaySurface((92, 192), (sfxH_Offset, 188 + (offset * 22), 64, 22))
				utilities.blitSignsToDisplaySurface((164, 192), (64 + musicH_Offset, 188, 64, 22))

				utilities.blitSignsToDisplaySurface((92, 192), (256, 188 + (offset * 22), 64, 22))

			else:
				utilities.blitSignsToDisplaySurface((92, 192), (sfxH_Offset, 188, 64, 22))
				utilities.blitSignsToDisplaySurface((164, 192), (64 + musicH_Offset, 188 + (offset * 22), 64, 22))

				utilities.blitSignsToDisplaySurface((164, 192), (256, 188 + (offset * 22), 64, 22))


		elif self.gameState == GAME_STATE_INGAME:
			utilities.showLevelBgSurface()

			self.pillManager.renderPills()
			self.bonusObjectManager.renderBonusObjects()
			self.renderUpdwardSmokePuffs()
			self.renderExplosions()
			self.renderPlayerShots()
			self.renderEnemyShots()
			self.renderEnemySpawners()
			self.renderEnemies()

			self.renderPlayer()

			utilities.showLevelFgSurface()

			utilities.drawLivesHeader()
			utilities.drawLives(self.lives)

			utilities.drawScoreHeader()
			utilities.drawScore(self.score)

			utilities.drawLevelNr(self.level)

			if self.getReadyTime == True:
				utilities.drawGetReadySign(self.getReadyFrameCounter)

			if self.isPausedGame == True:
				utilities.drawPausedSign(int(self.frameCounter60Hz / 30))

			if self.enterHiscoreMode == True:
				self.hiscoreManager.renderHiscoreEntryPanel(offset, self.score)

		elif self.gameState == GAME_STATE_PAUSED:
			pass

	#
	#
	#
	def storeGameDataForTermination(self):
		self.musicManager.saveSettings()
		self.soundFxManager.saveSettings()

	#
	#
	#
	def isGameActive(self):
		if self.isPausedGame == True:
			return False

		if self.getReadyTime == True or self.nextLevelTime == True:
			return False

		if self.playerState == PLAYER_STATE_DYING:
			return False

		return True


	#
	#
	#
	def prepareLevel(self):
		moduloLevel = ((self.level - 1) % 19) + 1
		self.levelFilename = 'VanSangsPain/levels/level' + str(moduloLevel)

		self.levelTiles = level.Level()
		self.levelTiles.load(self.levelFilename)

		self.spawnPoint = (-1, -1)
		self.buildSpawnCoordinate()

		self.pillManager.preparePills();
		self.bonusObjectManager.prepareBonusObjects();
		self.prepareUpdwardSmokePuffs();
		self.prepareExplosions();
		self.preparePlayerShots();
		self.prepareEnemyShots();
		self.prepareEnemySpawners()
		self.prepareEnemies()

		self.buildLevel()

		self.preparePlayerSprites()

		self.resetPlayerKeys()

		self.playerPosition = [(self.spawnPoint[0] * utilities.TILE_WIDTH) - 4, (self.spawnPoint[1] * utilities.TILE_HEIGHT) - 8]
		self.playerLastFacing = FACING_RIGHT
		self.changePlayerState(PLAYER_STATE_IDLE_RIGHT, True)

		self.playerDyingOffsetY = -4.0

	#
	#
	#
	def resetPlayerKeys(self):
		self.keyMoveLeftDown = False
		self.keyMoveRightDown = False
		self.keyFireDown = False
		self.keyJumpFloatDown = False
		self.keyDropDown = False

	#
	#
	#
	def keyMoveLeftStart(self):
		self.keyMoveLeftDown = True

		if self.playerState == PLAYER_STATE_IDLE_LEFT or self.playerState == PLAYER_STATE_IDLE_RIGHT:
			self.changePlayerState(PLAYER_STATE_WALK_LEFT, True)
			self.playerLastFacing = FACING_LEFT
		elif self.playerState == PLAYER_STATE_FLOAT_LEFT or self.playerState == PLAYER_STATE_FLOAT_RIGHT:
			self.changePlayerState(PLAYER_STATE_FLOAT_LEFT, True)
			self.playerLastFacing = FACING_LEFT
	#
	#
	#
	def keyMoveLeftEnd(self):
		self.keyMoveLeftDown = False

		if self.playerState == PLAYER_STATE_WALK_LEFT:
			self.changePlayerState(PLAYER_STATE_IDLE_LEFT, True)

	#
	#
	#
	def keyMoveRightStart(self):
		self.keyMoveRightDown = True

		if self.playerState == PLAYER_STATE_IDLE_LEFT or self.playerState == PLAYER_STATE_IDLE_RIGHT:
			self.changePlayerState(PLAYER_STATE_WALK_RIGHT, True)
			self.playerLastFacing = FACING_RIGHT
		elif self.playerState == PLAYER_STATE_FLOAT_LEFT or self.playerState == PLAYER_STATE_FLOAT_RIGHT:
			self.changePlayerState(PLAYER_STATE_FLOAT_RIGHT, True)
			self.playerLastFacing = FACING_RIGHT

	#
	#
	#
	def keyMoveRightEnd(self):
		self.keyMoveRightDown = False

		if self.playerState == PLAYER_STATE_WALK_RIGHT:
			self.changePlayerState(PLAYER_STATE_IDLE_RIGHT, True)

	#
	#
	#
	def keyJumpFloatStart(self):
		self.keyJumpFloatDown = True

		if self.playerState == PLAYER_STATE_IDLE_LEFT or self.playerState == PLAYER_STATE_IDLE_RIGHT or self.playerState == PLAYER_STATE_WALK_LEFT or self.playerState == PLAYER_STATE_WALK_RIGHT:
			self.changePlayerState(PLAYER_STATE_JUMPING, True)

			self.addUpdwardSmokePuff((self.playerPosition[0], self.playerPosition[1] + 8))

			self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_JUMP, 0)

	#
	#
	#
	def keyJumpFloatEnd(self):
		self.keyJumpFloatDown = False

		if self.playerState == PLAYER_STATE_JUMPING:
			if self.playerLastFacing == FACING_LEFT:
				self.changePlayerState(PLAYER_STATE_FLOAT_LEFT, True)
			else:
				self.changePlayerState(PLAYER_STATE_FLOAT_RIGHT, True)

			# ensure first player update leaves player at initial position
			self.playerPosition[1] += PLAYER_JUMP_SPEED_PER_FRAME

	#
	#
	#
	def keyDropStart(self):
		self.keyDropDown = True

	#
	#
	#
	def keyDropEnd(self):
		self.keyDropDown = False

	#
	#
	#
	def keyFireStart(self):
		self.keyFireDown = True

		if self.playerLastFacing == FACING_LEFT:
			shotSpawnPosition = (self.playerPosition[0] - 8 + PLAYER_SHOT_SPEED, self.playerPosition[1] + 8)
		else:
			shotSpawnPosition = (self.playerPosition[0] + 24 - PLAYER_SHOT_SPEED, self.playerPosition[1] + 8)

	#
	#
	#
	def keyFireEnd(self):
		self.keyFireDown = False

	#
	#
	#
	def movePlayer(self):
		if self.playerState == PLAYER_STATE_DYING and self.enterHiscoreMode == False:
			if self.playerPosition[1] >= 3600.0:
				self.lives -= 1
				if self.lives == 0:
					if self.score > self.hiscoreManager.getLowestHiscoreValue():
						self.enterHiscoreMode = True
						return
					else:
						self.enterGameState(GAME_STATE_HISCORES)
						return

				self.restartIngameLevel()
				return
			else:
				self.playerPosition[1] += self.playerDyingOffsetY
				self.playerDyingOffsetY += 0.25
				return

		if self.enterHiscoreMode == True:
			return

		if self.playerState == PLAYER_STATE_JUMPING:
			self.playerPosition[1] -= PLAYER_JUMP_SPEED_PER_FRAME

			headBlockStateCheck = self.checkPlayerHeadBlockedState()
			if(headBlockStateCheck[0] == True or headBlockStateCheck[1] == True):
				self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BOUNCE, 0)

				if self.playerLastFacing == FACING_LEFT:
					self.changePlayerState(PLAYER_STATE_FLOAT_LEFT, True)
				else:
					self.changePlayerState(PLAYER_STATE_FLOAT_RIGHT, True)

		else:
			canPlayerStand = self.checkPlayerStandableState()
 			leftPlayerBlock = self.checkPlayerLeftBlockedState()
 			rightPlayerBlock = self.checkPlayerRightBlockedState()

			if self.playerState == PLAYER_STATE_WALK_LEFT:
				if canPlayerStand == True:
					if leftPlayerBlock == False:
						self.playerPosition[0] -= PLAYER_HORZ_SPEED_PER_FRAME

				else:
					self.changePlayerState(PLAYER_STATE_FLOAT_LEFT, True)
					if self.keyMoveLeftDown == True and leftPlayerBlock == False:
						self.playerPosition[0] -= PLAYER_HORZ_SPEED_PER_FRAME

					if self.keyDropDown == True:
						self.playerPosition[1] += PLAYER_FALL_SPEED_PER_FRAME
					else:
						self.playerPosition[1] += PLAYER_FLOAT_SPEED_PER_FRAME

			elif self.playerState == PLAYER_STATE_WALK_RIGHT:
				if canPlayerStand == True:
					if rightPlayerBlock == False:
						self.playerPosition[0] += PLAYER_HORZ_SPEED_PER_FRAME

				else:
					self.changePlayerState(PLAYER_STATE_FLOAT_RIGHT, True)
					if self.keyMoveRightDown == True and rightPlayerBlock == False:
						self.playerPosition[0] += PLAYER_HORZ_SPEED_PER_FRAME

					if self.keyDropDown == True:
						self.playerPosition[1] += PLAYER_FALL_SPEED_PER_FRAME
					else:
						self.playerPosition[1] += PLAYER_FLOAT_SPEED_PER_FRAME

			elif self.playerState == PLAYER_STATE_FLOAT_LEFT or self.playerState == PLAYER_STATE_FLOAT_RIGHT:
				if canPlayerStand == False:
					if self.keyDropDown == True:
						self.playerPosition[1] += PLAYER_FALL_SPEED_PER_FRAME
					else:
						self.playerPosition[1] += PLAYER_FLOAT_SPEED_PER_FRAME

					if self.keyMoveLeftDown == True and leftPlayerBlock == False:
						self.playerPosition[0] -= PLAYER_HORZ_SPEED_PER_FRAME
					if self.keyMoveRightDown == True and rightPlayerBlock == False:
						self.playerPosition[0] += PLAYER_HORZ_SPEED_PER_FRAME
				else:
					self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BOUNCE, 0)

					if self.playerLastFacing == FACING_LEFT:
						if self.keyMoveLeftDown == True:
							self.changePlayerState(PLAYER_STATE_WALK_LEFT, True)
						else:
							self.changePlayerState(PLAYER_STATE_IDLE_LEFT, True)

					else:
						if self.keyMoveRightDown == True:
							self.changePlayerState(PLAYER_STATE_WALK_RIGHT, True)
						else:
							self.changePlayerState(PLAYER_STATE_IDLE_RIGHT, True)

	#
	#
	#
	def updatePlayer(self):
		self.playerAnimHoldFrameCounter += 1
		if self.playerAnimHoldFrameCounter >= PLAYER_ANIM_HOLD_DURATION:
			self.playerAnimHoldFrameCounter -= PLAYER_ANIM_HOLD_DURATION

			self.playerCurrAnimFrameNr += 1
			if self.playerState == PLAYER_STATE_WALK_LEFT:
				if self.playerCurrAnimFrameNr >= len(self.playerWalkLeftSpriteNrs):
					self.playerCurrAnimFrameNr = 0
			elif self.playerState == PLAYER_STATE_WALK_RIGHT:
				if self.playerCurrAnimFrameNr >= len(self.playerWalkRightSpriteNrs):
					self.playerCurrAnimFrameNr = 0
			elif self.playerState == PLAYER_STATE_JUMPING:
				if self.playerCurrAnimFrameNr >= len(self.playerJumpSpriteNrs):
					self.playerCurrAnimFrameNr = 0
			elif self.playerState == PLAYER_STATE_FLOAT_LEFT:
				if self.playerCurrAnimFrameNr >= len(self.playerFloatLeftSpriteNrs):
					self.playerCurrAnimFrameNr = 0
			elif self.playerState == PLAYER_STATE_FLOAT_RIGHT:
				if self.playerCurrAnimFrameNr >= len(self.playerFloatRightSpriteNrs):
					self.playerCurrAnimFrameNr = 0
			elif self.playerState == PLAYER_STATE_IDLE_LEFT:
				if self.playerCurrAnimFrameNr >= len(self.playerIdleLeftSpriteNrs):
					self.playerCurrAnimFrameNr = 0
			elif self.playerState == PLAYER_STATE_IDLE_RIGHT:
				if self.playerCurrAnimFrameNr >= len(self.playerIdleRightSpriteNrs):
					self.playerCurrAnimFrameNr = 0
			else: # probably dying
				if self.playerCurrAnimFrameNr >= len(self.playerDieSpriteNrs):
					self.playerCurrAnimFrameNr = 0

	#
	#
	#
	def renderPlayer(self):
		currentPlayerPosition = (self.playerPosition[0], self.playerPosition[1])

		if self.playerState == PLAYER_STATE_WALK_LEFT:
			currentSourceRect = self.playerWalkLeftSourceRects[self.playerCurrAnimFrameNr]
		elif self.playerState == PLAYER_STATE_WALK_RIGHT:
			currentSourceRect = self.playerWalkRightSourceRects[self.playerCurrAnimFrameNr]
		elif self.playerState == PLAYER_STATE_JUMPING:
			currentSourceRect = self.playerJumpSourceRects[self.playerCurrAnimFrameNr]
		elif self.playerState == PLAYER_STATE_FLOAT_LEFT:
			currentSourceRect = self.playerFloatLeftSourceRects[self.playerCurrAnimFrameNr]
		elif self.playerState == PLAYER_STATE_FLOAT_RIGHT:
			currentSourceRect = self.playerFloatRightSourceRects[self.playerCurrAnimFrameNr]
		elif self.playerState == PLAYER_STATE_IDLE_LEFT:
			currentSourceRect = self.playerIdleLeftSourceRects[self.playerCurrAnimFrameNr]
		elif self.playerState == PLAYER_STATE_IDLE_RIGHT:
			currentSourceRect = self.playerIdleRightSourceRects[self.playerCurrAnimFrameNr]
		else: # probably dying
			currentSourceRect = self.playerDieSourceRects[self.playerCurrAnimFrameNr]

		utilities.blitSprite24ToDisplaySurface(currentPlayerPosition, currentSourceRect)

	#
	#
	#
	def changePlayerState(self, newState, resetAnim = False):
		self.playerState = newState

		if resetAnim == True:
			self.playerCurrAnimFrameNr = 0

	#
	#
	#
	def checkPlayerStandableState(self):
		check1 = self.levelTiles.isMidgroundBlockedAtPixelPosition(self.playerPosition[0], self.playerPosition[1] + 23 + PLAYER_FALL_SPEED_PER_FRAME)
		check2 = self.levelTiles.isMidgroundBlockedAtPixelPosition(self.playerPosition[0] + 23, self.playerPosition[1] + 23 + PLAYER_FALL_SPEED_PER_FRAME)
		return check1 or check2

	#
	#
	#
	def checkPlayerHeadBlockedState(self):
		check1 = self.levelTiles.isMidgroundBlockedAtPixelPosition(self.playerPosition[0] + 7, self.playerPosition[1])
		check2 = self.levelTiles.isMidgroundBlockedAtPixelPosition(self.playerPosition[0] + 18, self.playerPosition[1])
		return (check1, check2)

	#
	#
	#
	def checkPlayerHipBlockedState(self):
		check1 = self.levelTiles.isMidgroundBlockedAtPixelPosition(self.playerPosition[0] + 7, self.playerPosition[1] + 12)
		check2 = self.levelTiles.isMidgroundBlockedAtPixelPosition(self.playerPosition[0] + 18, self.playerPosition[1] + 12)
		return (check1, check2)

	#
	#
	#
	def checkPlayerLeftBlockedState(self):
		check1 = self.levelTiles.isMidgroundBlockedAtPixelPosition(self.playerPosition[0] - PLAYER_HORZ_SPEED_PER_FRAME, self.playerPosition[1])
		check2 = self.levelTiles.isMidgroundBlockedAtPixelPosition(self.playerPosition[0] - PLAYER_HORZ_SPEED_PER_FRAME, self.playerPosition[1] + 12)
		check3 = self.levelTiles.isMidgroundBlockedAtPixelPosition(self.playerPosition[0] - PLAYER_HORZ_SPEED_PER_FRAME, self.playerPosition[1] + 23)
		return check1 or check2 or check3

	#
	#
	#
	def checkPlayerRightBlockedState(self):
		check1 = self.levelTiles.isMidgroundBlockedAtPixelPosition(self.playerPosition[0] + 23 + PLAYER_HORZ_SPEED_PER_FRAME, self.playerPosition[1])
		check2 = self.levelTiles.isMidgroundBlockedAtPixelPosition(self.playerPosition[0] + 23 + PLAYER_HORZ_SPEED_PER_FRAME, self.playerPosition[1] + 12)
		check3 = self.levelTiles.isMidgroundBlockedAtPixelPosition(self.playerPosition[0] + 23 + PLAYER_HORZ_SPEED_PER_FRAME, self.playerPosition[1] + 23)
		return check1 or check2 or check3

	#
	#
	#
	def preparePlayerSprites(self):
		self.playerAnimHoldFrameCounter = -1

		self.playerWalkLeftSpriteNrs = (0, 1, 2, 1)
		self.playerWalkLeftSourceRects = utilities.buildSprite24SourceRectsFromSpriteNrList(self.playerWalkLeftSpriteNrs)

		self.playerWalkRightSpriteNrs = (3, 4, 5, 4)
		self.playerWalkRightSourceRects = utilities.buildSprite24SourceRectsFromSpriteNrList(self.playerWalkRightSpriteNrs)

		self.playerJumpSpriteNrs = (10, 11, 12, 11)
		self.playerJumpSourceRects = utilities.buildSprite24SourceRectsFromSpriteNrList(self.playerJumpSpriteNrs)

		self.playerFloatLeftSpriteNrs = (13, 14, 15, 14)
		self.playerFloatLeftSourceRects = utilities.buildSprite24SourceRectsFromSpriteNrList(self.playerFloatLeftSpriteNrs)

		self.playerFloatRightSpriteNrs = (16, 17, 18, 17)
		self.playerFloatRightSourceRects = utilities.buildSprite24SourceRectsFromSpriteNrList(self.playerFloatRightSpriteNrs)

		self.playerIdleLeftSpriteNrs = (19, 20, 19, 21)
		self.playerIdleLeftSourceRects = utilities.buildSprite24SourceRectsFromSpriteNrList(self.playerIdleLeftSpriteNrs)

		self.playerIdleRightSpriteNrs = (24, 22, 24, 23)
		self.playerIdleRightSourceRects = utilities.buildSprite24SourceRectsFromSpriteNrList(self.playerIdleRightSpriteNrs)

		self.playerDieSpriteNrs = (6, 7, 8, 9)
		self.playerDieSourceRects = utilities.buildSprite24SourceRectsFromSpriteNrList(self.playerDieSpriteNrs)

	#
	#
	#
	def addLives(self, nos):
		self.lives += nos
		if self.lives > 99:
			self.lives = 99

	#
	#
	#
	def addScore(self, nos):
		self.score += nos
		if self.score >= 9999999:
			self.score = 9999999

	#
	#
	#
	def buildLevelBackground(self):
		# build BG
		utilities.clearLevelBgSurface()

		for row in range(0, utilities.TILE_ROWS) :
			for col in range (0, utilities.TILE_COLS) :
				targetPos = (col * utilities.TILE_WIDTH, row * utilities.TILE_HEIGHT)

				tileNr = self.levelTiles.getBackgroundTileNr(col, row)
				if tileNr != -1:
					sourceRect = pygame.Rect((tileNr % utilities.TILE_COLS) * utilities.TILE_WIDTH, (tileNr / utilities.TILE_COLS) *  utilities.TILE_HEIGHT, utilities.TILE_WIDTH,  utilities.TILE_HEIGHT)
					utilities.blitTileToLevelBgSurface(targetPos, sourceRect)

		# build MG
		for row in range(0, utilities.TILE_ROWS) :
			for col in range (0, utilities.TILE_COLS) :
				targetPos = (col * utilities.TILE_WIDTH, row * utilities.TILE_HEIGHT)

				tileNr = self.levelTiles.getMidgroundTileNr(col, row)
				if tileNr != -1:
					sourceRect = pygame.Rect((tileNr % utilities.TILE_COLS) * utilities.TILE_WIDTH, (tileNr / utilities.TILE_COLS) *  utilities.TILE_HEIGHT, utilities.TILE_WIDTH,  utilities.TILE_HEIGHT)
					utilities.blitTileToLevelBgSurface(targetPos, sourceRect)

	#
	#
	#
	def buildLevelForeground(self):
		# build FG
		utilities.clearLevelFgSurface()

		for row in range(0, utilities.TILE_ROWS) :
			for col in range (0, utilities.TILE_COLS) :
				targetPos = (col * utilities.TILE_WIDTH, row * utilities.TILE_HEIGHT)

				tileNr = self.levelTiles.getForegroundTileNr(col, row)
				if tileNr != -1:
					sourceRect = pygame.Rect((tileNr % utilities.TILE_COLS) * utilities.TILE_WIDTH, (tileNr / utilities.TILE_COLS) *  utilities.TILE_HEIGHT, utilities.TILE_WIDTH,  utilities.TILE_HEIGHT)
					utilities.blitTileToLevelFgSurface(targetPos, sourceRect)

	#
	#
	#
	def buildLevel(self):
		self.buildLevelBackground()
		self.buildLevelForeground()

	#
	#
	#
	def buildSpawnCoordinate(self):
		self.spawnPoint = self.levelTiles.findFirstTileNrInMidground(142)
		if self.spawnPoint != None:
			self.levelTiles.setMidgroundTileNr(self.spawnPoint[0], self.spawnPoint[1], -1)


	#
	#
	#
	def prepareUpdwardSmokePuffs(self):
		self.animatedUpwardPuffSmokeLifetimeList = []

	#
	#
	#
	def addUpdwardSmokePuff(self, (pixelX, pixelY)):
		puffSpriteNrs = (40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50)

		holdDuration = UPWARD_SMOKE_PUFF_HOLD_DURATION
		lifetime = len(puffSpriteNrs) * holdDuration
		doAnimate = True

		newUpwardSmokePuffAnimatedSprite = animatedSprite16.AnimatedSprite16((pixelX, pixelY), puffSpriteNrs, holdDuration, doAnimate)
		self.animatedUpwardPuffSmokeLifetimeList.append([lifetime, newUpwardSmokePuffAnimatedSprite])

	#
	#
	#
	def updateUpdwardSmokePuffs(self):
		#if self.isGameActive() == True:
		for index in range(len(self.animatedUpwardPuffSmokeLifetimeList) - 1, -1, -1):
			self.animatedUpwardPuffSmokeLifetimeList[index][0] -= 1
			if self.animatedUpwardPuffSmokeLifetimeList[index][0] <= 0:
				del	self.animatedUpwardPuffSmokeLifetimeList[index]

		for upwardSmokePuffCompound in self.animatedUpwardPuffSmokeLifetimeList:
			upwardSmokePuffCompound[1].update()

	#
	#
	#
	def renderUpdwardSmokePuffs(self):
		for upwardSmokePuffCompound in self.animatedUpwardPuffSmokeLifetimeList:
			upwardSmokePuffCompound[1].render()

	#
	#
	#
	def prepareExplosions(self):
		self.animatedExplosionsLifetimeList = []

	#
	#
	#
	def addExplosion(self, (pixelX, pixelY)):
		explosionSpriteNrs = (30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45)

		holdDuration = EXPLOSION_HOLD_DURATION
		lifetime = len(explosionSpriteNrs) * holdDuration
		doAnimate = True

		newExplosionAnimatedSprite = animatedSprite24.AnimatedSprite24((pixelX, pixelY), explosionSpriteNrs, holdDuration, True)
		self.animatedExplosionsLifetimeList.append([lifetime, newExplosionAnimatedSprite])

		self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_HIT_HARD, 0)

	#
	#
	#
	def updateExplosions(self):
		#if self.isGameActive() == True: # allow update during pause anyway since animation repeating is ugly otherwise
		for index in range(len(self.animatedExplosionsLifetimeList) - 1, -1, -1):
			self.animatedExplosionsLifetimeList[index][0] -= 1
			if self.animatedExplosionsLifetimeList[index][0] <= 0:
				del	self.animatedExplosionsLifetimeList[index]

		for explosionCompound in self.animatedExplosionsLifetimeList:
			explosionCompound[1].update()

	#
	#
	#
	def renderExplosions(self):
		for explosionCompound in self.animatedExplosionsLifetimeList:
			explosionCompound[1].render()

	#
	#
	#
	def preparePlayerShots(self):
		self.playerShotsCompoundList = []

	#
	#
	#
	def addPlayerShot(self, (pixelX, pixelY), facing):
		if self.playerShotIntervalCounter < (PLAYER_SHOT_INTERVAL / self.playerShotIntervalModifier):
			return
		else:
			self.playerShotIntervalCounter = 0

		playerShotsSpriteNrs = (1, 2, 3, 2)

		holdDuration = PLAYER_SHOTS_HOLD_DURATION
		lifetime = PLAYER_SHOTS_LIFETIME * self.playerShotLifetimeModifier
		doAnimate = True

		newPlayerShotSprite = animatedSprite8.AnimatedSprite8((pixelX, pixelY), playerShotsSpriteNrs, holdDuration, True)
		self.playerShotsCompoundList.append([lifetime, newPlayerShotSprite, facing])

		self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_TURRET_SHOT, 0)

	#
	#
	#
	def updatePlayerShots(self):
		if self.isGameActive() == True:
			self.playerShotIntervalCounter += 1

			for index in range(len(self.playerShotsCompoundList) - 1, -1, -1):
				self.playerShotsCompoundList[index][0] -= 1
				if self.playerShotsCompoundList[index][0] <= 0:
					oldPosition = self.playerShotsCompoundList[index][1].getPosition()
					self.addExplosion((oldPosition[0] + 4 - 12, oldPosition[1] + 4 - 12))

					del	self.playerShotsCompoundList[index]
				else:
					if self.playerShotsCompoundList[index][2] == FACING_LEFT:
						shotOffset = -(PLAYER_SHOT_SPEED * self.playerShotSpeedModifier)

					else:
						shotOffset = (PLAYER_SHOT_SPEED * self.playerShotSpeedModifier)

					oldPosition = self.playerShotsCompoundList[index][1].getPosition()
					newPosition = (oldPosition[0] + shotOffset, oldPosition[1])
					self.playerShotsCompoundList[index][1].setPosition(newPosition[0], newPosition[1])

					if self.checkPlayerShotAgainstEnemies(oldPosition, newPosition) == True:
						self.playerShotsCompoundList[index][0] = 0

					elif self.pillManager.checkPlayerShotAgainstPills(oldPosition, newPosition) == True:
						self.playerShotsCompoundList[index][0] = 0

					elif self.levelTiles.isMidgroundBlockedAtPixelPosition(newPosition[0] + 4, newPosition[1] + 4) == True:
						self.addExplosion((newPosition[0] + 4 - 12, newPosition[1] + 4 - 12))
						self.playerShotsCompoundList[index][0] = 0


		for shotCompound in self.playerShotsCompoundList:
			shotCompound[1].update()

	#
	#
	#
	def renderPlayerShots(self):
		if self.isGameActive() == True:
			for shotCompound in self.playerShotsCompoundList:
				shotCompound[1].render()

	#
	#
	#
	def checkPlayerShotAgainstEnemies(self, sweepStartPosition, sweepEndPosition):
		shotVector = (sweepEndPosition[0] - sweepStartPosition[0], sweepEndPosition[0] - sweepStartPosition[0])
		shotVectorLength = math.sqrt(math.pow(shotVector[0], 2) + math.pow(shotVector[1], 2))

		shotVectorUnitLength = (shotVector[0] / shotVectorLength, shotVector[1] / shotVectorLength);

		for shotVectorStep in range(0, int(shotVectorLength), 12):
			targetCoordinate = (sweepStartPosition[0] + (shotVectorUnitLength[0] * shotVectorStep), sweepStartPosition[1] + (shotVectorUnitLength[1] * shotVectorStep))

			for index in range(len(self.enemyList) - 1, -1, -1):
				enemyCompound = self.enemyList[index]

				enemyPosition = enemyCompound[1].getPosition()
				enemyCenter = (enemyPosition[0] + 12, enemyPosition[1] + 12)

				if utilities.distanceBetweenFloatCoordinates(targetCoordinate, enemyCenter) < ENEMY_HIT_DISTANCE:
					del self.enemyList[index]
					return True

		return False

	#
	#
	#
	def markRestartLevel(self):
		self.restartLevel = True

		self.resetPlayerKeys()

	#
	#
	#
	def markMoveToNextLevel(self):
		self.moveLevelUp = True

		self.nextLevelTime = True
		self.nextLevelFrameCounter = 0

		if self.playerLastFacing == FACING_LEFT:
			self.changePlayerState(PLAYER_STATE_IDLE_LEFT, True)
		else:
			self.changePlayerState(PLAYER_STATE_IDLE_RIGHT, True)

		self.resetPlayerKeys()

		voiceRandomIndex = random.randint(0, 3)
		if voiceRandomIndex == 0:
			self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_YAY, 1)
		elif voiceRandomIndex == 1:
			self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_YIHA, 1)
		elif voiceRandomIndex == 2:
			self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_HURRAY, 1)
		else:
			self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_LOADING_NEXT_LEVEL, 1)

	#
	#
	#
	def restartIngameLevel(self):
		self.moveLevelUp = False
		self.moveLevelDown = False
		self.startNewLevel = True

		self.getReadyTime = False
		self.getReadyFrameCounter = 0

		self.nextLevelTime = False
		self.nextLevelFrameCounter = 0

		self.isPausedGame = False

	#
	#
	#
	def moveToNextLevel(self):
		self.moveLevelUp = False

		self.level += 1
		if self.level > 99999999:
			self.level = 1

		self.startNewLevel = True

		#print("level up:")
		#print(self.level)

		pygame.time.wait(LEVEL_CHANGE_WAIT_TIME)

	#
	#
	#
	def moveToPrevLevel(self):
		self.moveLevelDown = False

		self.level -= 1
		if self.level < 1:
			self.level = 1

		self.startNewLevel = True

		#print("level down:")
		#print(self.level)

		pygame.time.wait(LEVEL_CHANGE_WAIT_TIME)

	#
	#
	#
	def prepareEnemySpawners(self):
		self.enemySpawnerList = []

		self.enemySpawnCounter = 0
		self.enemySpawnRate = ENEMY_SPAWN_FREQUENCY / (1.0 + (ENEMY_SPAWN_FACTOR * (self.level - 1)))
		if self.enemySpawnRate < ENEMY_SPAWN_MIN_FREQUENCY:
			self.enemySpawnRate < ENEMY_SPAWN_MIN_FREQUENCY
		#print("Enemy spawn rate: " + str(self.enemySpawnRate))

	#
	#
	#
	def addEnemySpawner(self, (pixelX, pixelY)):
		enemySpawnerSpriteNrs = (46, 47, 48, 49)

		holdDuration = ENEMY_SPAWNER_HOLD_DURATION
		lifetime = ENEMY_SPAWNER_LIFETIME

		newEnemySpawnSprite = animatedSprite24.AnimatedSprite24((pixelX, pixelY), enemySpawnerSpriteNrs, holdDuration, True)
		self.enemySpawnerList.append([lifetime, newEnemySpawnSprite])

		self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_SHOT_A, 0)

	#
	#
	#
	def updateEnemySpawners(self):
		if self.isGameActive() == True:
			if len(self.enemyList) < MAX_ENEMIES_ALIVE:
				self.enemySpawnCounter += 1
				if self.enemySpawnCounter >= self.enemySpawnRate:
					self.enemySpawnCounter = 0

					tiledPosition = self.findFirstNonBlockedTileInLevel()
					newPosition = ((tiledPosition[0] * utilities.TILE_WIDTH) - 4, (tiledPosition[1] * utilities.TILE_HEIGHT) - 4)
					if newPosition[0] != -1 and newPosition[1] != -1:
						#print("Adding enemy spawner X:"+ str(newPosition[0]) + ", Y:" + str(newPosition[1]))
						self.addEnemySpawner(newPosition)

			for index in range(len(self.enemySpawnerList) - 1, -1, -1):
				if self.isGameActive() == True:
					self.enemySpawnerList[index][0] -= 1
					if self.enemySpawnerList[index][0] <= 0:
						facingIndex = random.randint(0, 1)
						if facingIndex == 0:
							facing = FACING_LEFT
						else:
							facing = FACING_RIGHT

						self.spawnBestFitEnemyAtLocation(self.enemySpawnerList[index][1].getPosition(), facing)

						del self.enemySpawnerList[index]

		for enemySpawnCompound in self.enemySpawnerList:
			enemySpawnCompound[1].update()

	#
	#
	#
	def renderEnemySpawners(self):
		for enemySpawnCompound in self.enemySpawnerList:
			enemySpawnCompound[1].render()

	#
	#
	#
	def findFirstNonBlockedTileInLevel(self):
		nosTries = 1000

		while nosTries > 0:
			nosTries -= 1

			tileX = random.randint(1, utilities.TILE_COLS - 1)
			tileY = random.randint(1, utilities.TILE_ROWS - 1)

			check = self.levelTiles.isMidgroundBlockedAtPixelPosition(tileX * utilities.TILE_WIDTH, tileY * utilities.TILE_HEIGHT)
			if check == False:
				return (tileX, tileY)

		return (-1, -1)

	#
	#
	#
	def spawnBestFitEnemyAtLocation(self, position, facing):
		tilePixelCenterPosition = (position[0] + 12, position[1] + 12)

		origTileX = position[0] / utilities.TILE_WIDTH
		origTileY = position[1] / utilities.TILE_HEIGHT

		blockedTL = check = self.levelTiles.isMidgroundBlockedAtPixelPosition(tilePixelCenterPosition[0] - utilities.TILE_WIDTH, tilePixelCenterPosition[1] - utilities.TILE_HEIGHT)
		blockedTP = check = self.levelTiles.isMidgroundBlockedAtPixelPosition(tilePixelCenterPosition[0], tilePixelCenterPosition[1] - utilities.TILE_HEIGHT)
		blockedTR = check = self.levelTiles.isMidgroundBlockedAtPixelPosition(tilePixelCenterPosition[0] + utilities.TILE_WIDTH, tilePixelCenterPosition[1] - utilities.TILE_HEIGHT)

		#blockedLT = check = self.levelTiles.isMidgroundBlockedAtPixelPosition(tilePixelCenterPosition[0] - utilities.TILE_WIDTH, tilePixelCenterPosition[1])
		#blockedRT = check = self.levelTiles.isMidgroundBlockedAtPixelPosition(tilePixelCenterPosition[0] + utilities.TILE_WIDTH, tilePixelCenterPosition[1])

		blockedBL = check = self.levelTiles.isMidgroundBlockedAtPixelPosition(tilePixelCenterPosition[0] - utilities.TILE_WIDTH, tilePixelCenterPosition[1] + utilities.TILE_HEIGHT)
		blockedBT= check = self.levelTiles.isMidgroundBlockedAtPixelPosition(tilePixelCenterPosition[0], tilePixelCenterPosition[1] + utilities.TILE_HEIGHT)
		blockedBR = check = self.levelTiles.isMidgroundBlockedAtPixelPosition(tilePixelCenterPosition[0] + utilities.TILE_WIDTH, tilePixelCenterPosition[1] + utilities.TILE_HEIGHT)

		if blockedTP == True and blockedTL == True and blockedTR == True:
			modPosition = (position[0], position[1] + 4)
			self.addEnemy(ENEMY_TYPE_CEILING_GUN_TURRET, modPosition, facing)
			return

		elif blockedBT == True:
			if blockedBL == True and blockedBR == True:
				modPosition = (position[0], position[1] - 4)
				self.addEnemy(ENEMY_TYPE_TANK, modPosition, facing)
				return

			else:
				modPosition = (position[0], position[1] - 4)
				self.addEnemy(ENEMY_TYPE_FLOOR_GUN_TURRET, modPosition, facing)
				return

		#default enemy to create if all else filters through
		enemyChoiceIndex = random.randint(0, 1)
		if enemyChoiceIndex == 0:
			modPosition = (position[0], position[1])
			self.addEnemy(ENEMY_TYPE_HOMING_MISSILE, modPosition, facing)
		else:
			modPosition = (position[0], position[1])
			self.addEnemy(ENEMY_TYPE_SAW_BLADE, modPosition, facing)

	#
	#
	#
	def prepareEnemies(self):
		self.enemyList = []

		self.ModifiedHomingMissileSpeed = ENEMY_HOMING_MISSILE_SPEED + (self.level * 0.025)
		if self.ModifiedHomingMissileSpeed > 0.7:
			self.ModifiedHomingMissileSpeed = 0.7

		self.ModifiedSawBladeSpeed = ENEMY_SAW_BLADE_SPEED + (self.level * 0.025)
		if self.ModifiedSawBladeSpeed > 1.0:
			self.ModifiedSawBladeSpeed = 1.0

		self.ModifiedTankSpeed = ENEMY_TANK_SPEED + (self.level * 0.05)
		if self.ModifiedTankSpeed > 0.7:
			self.ModifiedTankSpeed = 0.7

		self.modifiedEnemySensorRange = ENEMY_SENSOR_RANGE + (self.level * (0.1 * 16.0))

	#
	#
	#
	def addEnemy(self, enemyType, position, facing):
		velocity = [0.0, 0.0]

		if enemyType == ENEMY_TYPE_HOMING_MISSILE:
			enemySpriteNrsFacingLeft = (50,)
			enemySpriteNrsFacingRight = (51,)
			holdDuration = ENEMY_HOMING_MISSILE_HOLD_DURATION

		elif enemyType == ENEMY_TYPE_SAW_BLADE:
			enemySpriteNrsFacingLeft = (56, 57, 58, 59)
			enemySpriteNrsFacingRight = (52, 53, 54, 55)
			holdDuration = ENEMY_SAW_BLADE_HOLD_DURATION

			velocityRandomIndex = random.randint(0, 1)
			if facing == FACING_LEFT:
				if velocityRandomIndex == 0:
					velocity = [-1.0, -1.0]
				else:
					velocity = [-1.0, 1.0]
			else:
				if velocityRandomIndex == 0:
					velocity = [1.0, -1.0]
				else:
					velocity = [1.0, 1.0]

		elif enemyType == ENEMY_TYPE_FLOOR_GUN_TURRET:
			enemySpriteNrsFacingLeft = (80, 69, 81)
			enemySpriteNrsFacingRight = (80, 69, 81)
			holdDuration = ENEMY_FLOOR_GUN_TURRET_HOLD_DURATION

		elif enemyType == ENEMY_TYPE_CEILING_GUN_TURRET:
			enemySpriteNrsFacingLeft = (90, 79, 91)
			enemySpriteNrsFacingRight = (90, 79, 91)
			holdDuration = ENEMY_CEILING_GUN_TURRET_HOLD_DURATION

		else: # must be a tank
			enemySpriteNrsFacingLeft = (82, 83, 86, 87)
			enemySpriteNrsFacingRight = (84, 85, 88, 89)
			holdDuration = ENEMY_TANK_HOLD_DURATION

		newEnemyLeftSprite = animatedSprite24.AnimatedSprite24((position[0], position[1]), enemySpriteNrsFacingLeft, holdDuration, True)
		newEnemyRightSprite = animatedSprite24.AnimatedSprite24((position[0], position[1]), enemySpriteNrsFacingRight, holdDuration, True)

		self.enemyList.append([enemyType, newEnemyLeftSprite, newEnemyRightSprite, facing, velocity])

		self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_SHOT_B, 0)

	#
	#
	#
	def updateEnemies(self):
		if self.isGameActive() == True:
			# move enemy according to type
			playerCenterPosition = (self.playerPosition[0] + 12, self.playerPosition[1] + 12)

			for enemyCompoundIndex in range(len(self.enemyList) - 1, -1, -1):
				enemyCompound = self.enemyList[enemyCompoundIndex]

				enemyType = enemyCompound[0]

				enemyPosition = enemyCompound[1].getPosition()
				enemyCenterPosition = (enemyPosition[0] + 12, enemyPosition[1] + 12)

				sensorRangeRequired = utilities.distanceBetweenFloatCoordinates(playerCenterPosition, enemyCenterPosition)

				if sensorRangeRequired <= ENEMY_DETONATE_DISTANCE and (enemyType == ENEMY_TYPE_HOMING_MISSILE or enemyType == ENEMY_TYPE_SAW_BLADE):
					self.addExplosion(self.playerPosition)
					del self.enemyList[enemyCompoundIndex]
					self.changePlayerState(PLAYER_STATE_DYING) # player caught in explosion!
					self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_OH_NO, 1)

				else:
					# player-position dependent processing
					if self.playerPosition[0] < enemyPosition[0]:
						enemyCompound[3] = FACING_LEFT

						if enemyType == ENEMY_TYPE_HOMING_MISSILE:
							blockedLU = self.levelTiles.isMidgroundBlockedAtPixelPosition(enemyPosition[0] - self.ModifiedHomingMissileSpeed, enemyPosition[1])
							blockedLM = self.levelTiles.isMidgroundBlockedAtPixelPosition(enemyPosition[0] - self.ModifiedHomingMissileSpeed, enemyPosition[1] + 12)
							blockedLD = self.levelTiles.isMidgroundBlockedAtPixelPosition(enemyPosition[0] - self.ModifiedHomingMissileSpeed, enemyPosition[1] + 23)

							if blockedLU == False and blockedLM == False and blockedLD == False:
								newEnemyPosition = (enemyPosition[0] - self.ModifiedHomingMissileSpeed, enemyPosition[1])
								enemyCompound[1].setPosition(newEnemyPosition[0], newEnemyPosition[1])
								enemyCompound[2].setPosition(newEnemyPosition[0], newEnemyPosition[1])
							else:
								self.addExplosion(enemyPosition)
								del self.enemyList[enemyCompoundIndex]
								continue

						elif enemyType == ENEMY_TYPE_SAW_BLADE:
							pass

						elif enemyType == ENEMY_TYPE_FLOOR_GUN_TURRET:
							pass

						elif enemyType == ENEMY_TYPE_CEILING_GUN_TURRET:
							pass

						else: # must be a tank
							if sensorRangeRequired <= self.modifiedEnemySensorRange and sensorRangeRequired >= utilities.SPRITE24_WIDTH:
								blockedLM = self.levelTiles.isMidgroundBlockedAtPixelPosition(enemyPosition[0] - self.ModifiedTankSpeed, enemyPosition[1] + 12)
								blockedGND = self.levelTiles.isMidgroundBlockedAtPixelPosition(enemyPosition[0] - self.ModifiedTankSpeed, enemyPosition[1] + 24)

								if blockedLM == False and blockedGND == True:
									newEnemyPosition = (enemyPosition[0] - self.ModifiedTankSpeed, enemyPosition[1])
									enemyCompound[1].setPosition(newEnemyPosition[0], newEnemyPosition[1])
									enemyCompound[2].setPosition(newEnemyPosition[0], newEnemyPosition[1])

					elif self.playerPosition[0] > enemyPosition[0]:
						enemyCompound[3] = FACING_RIGHT

						if enemyType == ENEMY_TYPE_HOMING_MISSILE:
							blockedRU = self.levelTiles.isMidgroundBlockedAtPixelPosition(enemyPosition[0] + 23 + self.ModifiedHomingMissileSpeed, enemyPosition[1])
							blockedRM = self.levelTiles.isMidgroundBlockedAtPixelPosition(enemyPosition[0] + 23 +  self.ModifiedHomingMissileSpeed, enemyPosition[1] + 12)
							blockedRD = self.levelTiles.isMidgroundBlockedAtPixelPosition(enemyPosition[0] + 23 +  self.ModifiedHomingMissileSpeed, enemyPosition[1] + 23)

							if blockedRU == False and blockedRM == False and blockedRD == False:
								newEnemyPosition = (enemyPosition[0] + self.ModifiedHomingMissileSpeed, enemyPosition[1])
								enemyCompound[1].setPosition(newEnemyPosition[0], newEnemyPosition[1])
								enemyCompound[2].setPosition(newEnemyPosition[0], newEnemyPosition[1])
							else:
								self.addExplosion(enemyPosition)
								del self.enemyList[enemyCompoundIndex]
								continue

						elif enemyType == ENEMY_TYPE_SAW_BLADE:
							pass

						elif enemyType == ENEMY_TYPE_FLOOR_GUN_TURRET:
							pass

						elif enemyType == ENEMY_TYPE_CEILING_GUN_TURRET:
							pass

						else: # must be a tank
							if sensorRangeRequired <= self.modifiedEnemySensorRange and sensorRangeRequired >= utilities.SPRITE24_WIDTH:
								blockedRM = self.levelTiles.isMidgroundBlockedAtPixelPosition(enemyPosition[0] + 23  + self.ModifiedTankSpeed, enemyPosition[1] + 12)
								blockedGND = self.levelTiles.isMidgroundBlockedAtPixelPosition(enemyPosition[0] + 23 + self.ModifiedTankSpeed, enemyPosition[1] + 24)

								if blockedRM == False and blockedGND == True:
									newEnemyPosition = (enemyPosition[0] + self.ModifiedTankSpeed, enemyPosition[1])
									enemyCompound[1].setPosition(newEnemyPosition[0], newEnemyPosition[1])
									enemyCompound[2].setPosition(newEnemyPosition[0], newEnemyPosition[1])

					enemyPosition = enemyCompound[1].getPosition()

					if self.playerPosition[1] < enemyPosition[1]:
						if enemyType == ENEMY_TYPE_HOMING_MISSILE:
							blockedUL =self.levelTiles.isMidgroundBlockedAtPixelPosition(enemyPosition[0], enemyPosition[1] - self.ModifiedHomingMissileSpeed)
							blockedUM = self.levelTiles.isMidgroundBlockedAtPixelPosition(enemyPosition[0] + 12, enemyPosition[1] - self.ModifiedHomingMissileSpeed)
							blockedUR = self.levelTiles.isMidgroundBlockedAtPixelPosition(enemyPosition[0] + 23, enemyPosition[1] - self.ModifiedHomingMissileSpeed)

							if blockedUL == False and blockedUM == False and blockedUR == False:
								newEnemyPosition = (enemyPosition[0], enemyPosition[1] - self.ModifiedHomingMissileSpeed)
								enemyCompound[1].setPosition(newEnemyPosition[0], newEnemyPosition[1])
								enemyCompound[2].setPosition(newEnemyPosition[0], newEnemyPosition[1])
							else:
								self.addExplosion(enemyPosition)
								del self.enemyList[enemyCompoundIndex]
								continue

						elif enemyType == ENEMY_TYPE_SAW_BLADE:
							pass

						elif enemyType == ENEMY_TYPE_FLOOR_GUN_TURRET:
							pass

						elif enemyType == ENEMY_TYPE_CEILING_GUN_TURRET:
							pass

						else: # must be a tank
							pass

					elif self.playerPosition[1] > enemyPosition[1]:
						if enemyType == ENEMY_TYPE_HOMING_MISSILE:
							blockedLL = self.levelTiles.isMidgroundBlockedAtPixelPosition(enemyPosition[0], enemyPosition[1] + 23 + self.ModifiedHomingMissileSpeed)
							blockedLM = self.levelTiles.isMidgroundBlockedAtPixelPosition(enemyPosition[0] + 12, enemyPosition[1] + 23 + self.ModifiedHomingMissileSpeed)
							blockedLR = self.levelTiles.isMidgroundBlockedAtPixelPosition(enemyPosition[0] + 23, enemyPosition[1] + 23 + self.ModifiedHomingMissileSpeed)

							if blockedLL == False and blockedLM == False and blockedLR == False:
								newEnemyPosition = (enemyPosition[0], enemyPosition[1] + self.ModifiedHomingMissileSpeed)
								enemyCompound[1].setPosition(newEnemyPosition[0], newEnemyPosition[1])
								enemyCompound[2].setPosition(newEnemyPosition[0], newEnemyPosition[1])
							else:
								self.addExplosion(enemyPosition)
								del self.enemyList[enemyCompoundIndex]
								continue

						elif enemyType == ENEMY_TYPE_SAW_BLADE:
							pass

						elif enemyType == ENEMY_TYPE_FLOOR_GUN_TURRET:
							pass

						elif enemyType == ENEMY_TYPE_CEILING_GUN_TURRET:
							pass

						else: # must be a tank
							pass

					# non-player-position dependent processing
					if enemyType == ENEMY_TYPE_SAW_BLADE:
						didBounce = False

						blockedLU = self.levelTiles.isMidgroundBlockedAtPixelPosition(enemyPosition[0] - self.ModifiedSawBladeSpeed, enemyPosition[1])
						blockedLD = self.levelTiles.isMidgroundBlockedAtPixelPosition(enemyPosition[0] - self.ModifiedSawBladeSpeed, enemyPosition[1] + 23)
						if blockedLU == True or blockedLD == True:
							enemyCompound[4][0] = 1.0
							enemyCompound[3] = FACING_RIGHT
							didBounce = True

						blockedRU = self.levelTiles.isMidgroundBlockedAtPixelPosition(enemyPosition[0] + 23 + self.ModifiedSawBladeSpeed, enemyPosition[1])
						blockedRD = self.levelTiles.isMidgroundBlockedAtPixelPosition(enemyPosition[0] + 23 + self.ModifiedSawBladeSpeed, enemyPosition[1] + 23)
						if blockedRU == True or blockedRD == True:
							enemyCompound[4][0] = -1.0
							enemyCompound[3] = FACING_LEFT
							didBounce = True

						blockedUL = self.levelTiles.isMidgroundBlockedAtPixelPosition(enemyPosition[0], enemyPosition[1] - self.ModifiedSawBladeSpeed)
						blockedUR = self.levelTiles.isMidgroundBlockedAtPixelPosition(enemyPosition[0] + 23, enemyPosition[1] - self.ModifiedSawBladeSpeed)
						if blockedUL == True or blockedUR == True:
							enemyCompound[4][1] = 1.0
							didBounce = True

						blockedDL = self.levelTiles.isMidgroundBlockedAtPixelPosition(enemyPosition[0], enemyPosition[1] + 23 + self.ModifiedSawBladeSpeed)
						blockedDR = self.levelTiles.isMidgroundBlockedAtPixelPosition(enemyPosition[0] + 23, enemyPosition[1] + 23 + self.ModifiedSawBladeSpeed)
						if blockedDL == True or blockedDR == True:
							enemyCompound[4][1] = -1.0
							didBounce = True

						if didBounce == True:
							self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_BOUNCE, 0)

						newEnemyPosition = (enemyPosition[0] + (enemyCompound[4][0] * self.ModifiedSawBladeSpeed), enemyPosition[1]  + (enemyCompound[4][1] * self.ModifiedSawBladeSpeed))
						enemyCompound[1].setPosition(newEnemyPosition[0], newEnemyPosition[1])
						enemyCompound[2].setPosition(newEnemyPosition[0], newEnemyPosition[1])

					elif enemyType == ENEMY_TYPE_CEILING_GUN_TURRET or enemyType == ENEMY_TYPE_FLOOR_GUN_TURRET or enemyType == ENEMY_TYPE_TANK:
						if sensorRangeRequired <= self.modifiedEnemySensorRange:
							if enemyCompound[3] == FACING_LEFT:
								self.addEnemyShot((enemyPosition[0], enemyPosition[1] + 12))
							else:
								self.addEnemyShot((enemyPosition[0] + 23, enemyPosition[1] + 12))

		for enemyCompound in self.enemyList:
			if enemyCompound[3] == FACING_LEFT:
				enemyCompound[1].update()
			else:
				enemyCompound[2].update()

	#
	#
	#
	def renderEnemies(self):
		for enemyCompound in self.enemyList:
			if enemyCompound[3] == FACING_LEFT:
				enemyCompound[1].render()
			else:
				enemyCompound[2].render()

	#
	#
	#
	def prepareEnemyShots(self):
		self.enemyShotsCompoundList = []

		self.enemyShotInterval = ENEMY_SHOT_INTERVAL - self.level
		if self.enemyShotInterval < ENEMY_MIN_SHOT_INTERVAL:
			self.enemyShotInterval = ENEMY_MIN_SHOT_INTERVAL

		self.enemyShotIntervalCounter = 0

		self.modifiedEnemyShotSpeed = ENEMY_SHOTS_SPEED + (0.1 * self.level)

	#
	#
	#
	def addEnemyShot(self, (pixelX, pixelY)):
		if self.enemyShotIntervalCounter < self.enemyShotInterval:
			return

		self.enemyShotIntervalCounter = 0

		enemyShotsSpriteNrs = (4, 5, 6, 5)

		holdDuration = ENEMY_SHOTS_HOLD_DURATION
		doAnimate = True

		playerCenterCoordinate = (self.playerPosition[0] + 12, self.playerPosition[1] + 12)
		shotCenterCoordinate = (pixelX, pixelY)

		vector = (playerCenterCoordinate[0] - shotCenterCoordinate[0], playerCenterCoordinate[1] - shotCenterCoordinate[1])
		vectorLength = utilities.distanceBetweenFloatCoordinates(playerCenterCoordinate, shotCenterCoordinate)
		unitDirectionVector = (vector[0] / vectorLength, vector[1] / vectorLength)

		newEnemyShotSprite = animatedSprite8.AnimatedSprite8(shotCenterCoordinate, enemyShotsSpriteNrs, holdDuration, True)
		self.enemyShotsCompoundList.append([newEnemyShotSprite, unitDirectionVector])

		self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_TURRET_SHOT, 0)

	#
	#
	#
	def updateEnemyShots(self):
		if self.isGameActive() == True:
			self.enemyShotIntervalCounter += 1

			for index in range(len(self.enemyShotsCompoundList) - 1, -1, -1):
					enemyShotCompound = self.enemyShotsCompoundList[index]

					shotOffset = (enemyShotCompound[1][0] * self.modifiedEnemyShotSpeed, enemyShotCompound[1][1] * self.modifiedEnemyShotSpeed)

					oldPosition = enemyShotCompound[0].getPosition()
					newPosition = (oldPosition[0] + shotOffset[0], oldPosition[1] + shotOffset[1])
					enemyShotCompound[0].setPosition(newPosition[0], newPosition[1])

					# check against player center
					playerCenterPosition = (self.playerPosition[0] + 12, self.playerPosition[1] + 12)
					shotCenterPosition = (newPosition[0] + 4, newPosition[1] + 4)
					if utilities.distanceBetweenFloatCoordinates(playerCenterPosition, shotCenterPosition) < PLAYER_HIT_DISTANCE:
						self.changePlayerState(PLAYER_STATE_DYING)
						self.soundFxManager.emitSoundEffect(self.soundFxManager.SFX_OH_NO, 1)

						self.addExplosion((newPosition[0] + 4 - 12, newPosition[1] + 4 - 12))
						del self.enemyShotsCompoundList[index]

					elif self.levelTiles.isMidgroundBlockedAtPixelPosition(newPosition[0] + 4, newPosition[1] + 4) == True:
						self.addExplosion((newPosition[0] + 4 - 12, newPosition[1] + 4 - 12))
						del self.enemyShotsCompoundList[index]

		for shotCompound in self.enemyShotsCompoundList:
			shotCompound[0].update()

	#
	#
	#
	def renderEnemyShots(self):
		if self.isGameActive() == True:
			for shotCompound in self.enemyShotsCompoundList:
				shotCompound[0].render()
